"""
Classes and functions to manage Web services authentication
"""

import base64

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.cache import cache
from django.http import HttpResponseForbidden
from django.utils.crypto import constant_time_compare
from django.utils.importlib import import_module

from functools import wraps


def import_by_path(dotted_path):
    """
    Reimplement this from django dev, can be replaced in Django 1.6

    Not as resiliant as the real version.
    """

    module_path, class_name = dotted_path.rsplit('.', 1)
    module = import_module(module_path)
    return getattr(module, class_name)


class Consumer(object):
    """
    Consumer class to supply to signature checking routines
    """

    def __init__(self, key=None, secret=None, obj=None):
        self._key = key
        self._secret = secret
        self._object = obj

    def key(self):
        """
        Return the consumer key
        """

        return self._key

    def secret(self):
        """
        Return the consumer secret
        """

        return self._secret


class ConsumerStore(object):
    """
    A default, database-backed store for looking up consumers by their API
    key.
    """

    consumer_class_key = 'key'
    consumer_class_secret = 'secret'

    @staticmethod
    def get_consumer_store():
        """
        Return the configured consumer store (not a singleton)
        """

        try:
            consumer_store_class = \
                import_by_path(settings.CONSUMER_STORE_CLASS)
        except AttributeError:
            raise ImproperlyConfigured(
                "Using CheckSignatureMiddleware requires "
                "settings.CONSUMER_STORE_CLASS")

        return consumer_store_class()

    @property
    def consumer_class(self):
        """
        The consumer class is a model in your database.

        The class must have a DB attribute suitable for use with a
        QuerySet.get() call, by default this is 'key' but can be overridden by
        defining consumer_class_key.

        The secret for the consumer is stored in the attribute 'secret' but
        can be overridden by defining consumer_class_secret.
        """

        raise ImproperlyConfigured("You must defined consumer_class")

    def get_consumer(self, key):
        """
        Retrieve the consumer from the store.
        """
        cache_key = 'consumer-%s' % key

        try:
            consumer = cache.get(cache_key)

            if isinstance(consumer, self.consumer_class.DoesNotExist):
                return None
            elif consumer is not None:
                return consumer

            obj = self.consumer_class.objects.get(
                **{self.consumer_class_key: key})

            consumer = Consumer(
                key=key,
                secret=getattr(obj, self.consumer_class_secret),
                obj=obj)

            cache.set(cache_key, consumer)
            return consumer

        except self.consumer_class.DoesNotExist as ex:

            cache.set(cache_key, ex)
            return None


class CheckSignatureMiddleware(object):
    """
    A middleware to check HTTP Basic and key parameter authentication.
    """

    def __init__(self):
        self.consumer_store = ConsumerStore.get_consumer_store()

    @staticmethod
    def get_basic_auth_headers(request):
        """
        Get HTTP Basic username and password from the request, if present.
        """

        try:
            (auth_type, data) = request.META['HTTP_AUTHORIZATION'].split()

            if auth_type.decode().lower() != 'basic':
                return None

            bits = base64.b64decode(data).decode('utf-8')
            (user, password) = bits.split(':')
            return (user, password)
        except:  # pylint:disable=bare-except
            return None

    def verify_consumer(self, key, secret):
        """
        Given a key,secret pair, either returns the asociated consumer,
        or return none.
        """

        consumer = self.consumer_store.get_consumer(key)
        if not consumer:
            return None

        if not constant_time_compare(consumer.secret(), secret):
            return None

        return consumer

    def http_basic_consumer(self, request):
        """
        If the request has a valid HTTP Basic authentication, return the
        associated consumer; None otherwise.
        """

        http_basic = self.get_basic_auth_headers(request)
        if not http_basic:
            return None

        (key, secret) = http_basic  # pylint:disable=unpacking-non-sequence

        return self.verify_consumer(key, secret)

    def key_parameter_consumer(self, request):
        """
        If the request has a parameter named 'key', which consists of valid
        consumer credentials, the associated consumer will be returned;
        otherwise None.
        """

        try:
            (key, secret) = request.GET.get('key').split(':')
        except (AttributeError, ValueError):
            key = secret = None

        if not key and not secret:
            return None

        return self.verify_consumer(key, secret)

    def process_request(self, request):
        """
        Check HTTP Basic and key parameter authentication.
        """
        methods = (
            self.http_basic_consumer,
            self.key_parameter_consumer,
        )

        for method in methods:
            consumer = method(request)
            if consumer:
                request.consumer = consumer
                return


def consumer_required(view):
    """
    Decorator for views that checks that the request has a valid signature, and
    responds with Forbidden otherwise.
    """
    @wraps(view)
    # pylint:disable=C0111
    # Missing docstring
    def wrapper(request, *args, **kwargs):
        consumer = getattr(request, 'consumer', None)
        if consumer is None:
            return HttpResponseForbidden()
        return view(request, *args, **kwargs)
    return wrapper



"""
Model mixins for authorization.
"""

import random

from django.db import models


def random_key(length):
    """
    Generate a 'random' key of length 'length'

    We are using pseudorandom since we are not a bank.
    """

    key = ''
    for _ in range(length):
        key += random.choice('ABCDEFGHIJKLMNOPQRSTUVWXYZ')

    return key


def random_consumer_key():
    """
    A random key for consumer.
    """

    return random_key(32)


class Consumer(models.Model):
    """
    An API consumer.
    """

    class Meta(object):
        abstract = True

    key = models.CharField(max_length=32, default=random_consumer_key,
                           unique=True)
    secret = models.CharField(max_length=32, default=random_consumer_key)

    def __unicode__(self):
        return self.key



"""
Tastypie API authentication handlers.
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from future.builtins import *

from tastypie.authentication import Authentication
from tastypie.http import HttpUnauthorized


class ApplicationAuthentication(Authentication):
    """
    Authenticate the API request by checking the application key.
    """

    realm = 'tastypie'

    def is_authenticated(self, request, **kwargs):
        """
        Check that the request is signed by the application.
        """
        consumer = getattr(request, 'consumer', None)
        if consumer is None:
            response = HttpUnauthorized()
            response['WWW-Authenticate'] = 'Basic Realm="%s"' % self.realm
            return response
        else:
            return True

    def get_identifier(self, request):
        """
        Return a combination of the consumer, the IP address and the host
        """

        consumer = getattr(request, 'consumer', None)
        return '%s_%s' % (
            consumer.key(),
            super(ApplicationAuthentication, self).get_identifier(request))



"""
Django REST Framework authentication handlers.
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from future.builtins import *

from rest_framework import (
    authentication,
    exceptions,
)


class ApplicationAuthentication(authentication.BaseAuthentication):
    """Authenticate against the API keys."""

    www_authenticate_realm = 'api'

    def authenticate(self, request):
        """
        Check that the request is signed by the application.
        """

        consumer = getattr(request, 'consumer', None)
        if consumer is None:
            raise exceptions.AuthenticationFailed("No valid API key.")

        # Make consumer look like a user for DRF
        consumer.is_authenticated = lambda: True

        return (consumer, None)

    def authenticate_header(self, request):
        return 'Basic realm="%s"' % self.www_authenticate_realm



"""
Steps for testing authentication
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from future.builtins import *

import base64

from django.test.client import Client

try:
    from aloe import before, step, world
except ImportError:
    from lettuce import before, step, world

from ixwsauth_server.middleware import ConsumerStore


class ApplicationClient(Client):
    """
    A Django test client authenticating using the consumer store.
    """

    consumer_store = ConsumerStore.get_consumer_store()

    def __init__(self, key, secret=None, **defaults):
        self._key = key

        if secret:
            self._secret = secret
        else:
            consumer = self.consumer_store.get_consumer(key)
            self._secret = consumer.secret()

        super(ApplicationClient, self).__init__(**defaults)

    def authorisation(self, request):
        """
        The HTTP Authorization header to add.
        """

        raise NotImplementedError("Must override authorisation().")

    def _base_environ(self, **request):
        """
        Add the HTTP Authorization header to the request.
        """

        environ = super(ApplicationClient, self)._base_environ(**request)

        environ['HTTP_AUTHORIZATION'] = self.authorisation(request)

        return environ


class BasicAuthClient(ApplicationClient):
    """
    A Django test client which authenticates request using HTTP Basic auth
    and a consumer from the consumer store.
    """

    def authorisation(self, request):
        """
        The HTTP Basic authorisation to add to the request.
        """

        auth = '{key}:{secret}'.format(key=self._key, secret=self._secret)
        base64string = base64.encodestring(auth.encode())\
            .replace(bytes(b'\n'), bytes(b''))

        return bytes(b'Basic ') + base64string


class KeyParameterClient(ApplicationClient):
    """
    A Django test client which authenticates request using a 'key' parameter
    and a consumer from the consumer store.
    """

    def authorisation(self, request):
        """
        Adds the key,secret pair to the request.
        """

        request.GET['key'] = '{key}:{secret}'.format(
            key=self._key,
            secret=self._secret,
        )


# each_example is not available in Lettuce master
# pylint:disable=invalid-name,maybe-no-member
try:
    before_example = before.each_example
except AttributeError:
    before_example = before.each_scenario
# pylint:enable=invalid-name,maybe-no-member


@before_example
def set_default_client(scenario, outline, steps):
    """
    Set a default client that does not have authentication
    """

    world.client = Client()


@step(r'I authenticate to the API with key "([^\"]*)"$')
@step(r'I authenticate to the API using HTTP Basic auth '
      r'with key "([^\"]*)"$')
def authenticate_application_basic(step_, key):
    """
    Authenticate as the application with given key and corresponding secret,
    using HTTP Basic.
    """

    world.client = BasicAuthClient(key=key)


@step(r'I authenticate to the API with key "([^"]*)" and secret "([^"]*)"$')
@step(r'I authenticate to the API using HTTP Basic auth '
      r'with key "([^\"]*)" and secret "([^\"]*)"')
def authenticate_with_secret_basic(step_, key, secret):
    """
    Authenticate to the application with the given key and secret,
    using HTTP Basic.
    """

    world.client = BasicAuthClient(key=key, secret=secret)


@step(r'I authenticate to the API using a key parameter '
      r'with key "([^"]*)" and secret "([^"]*)"')
def authenticate_with_key_parameter(step_, key, secret):
    """
    Authenticate to the application with the given key and secret,
    using a key parameter.
    """

    world.client = KeyParameterClient(key=key, secret=secret)



"""
Package configuration
"""
# pylint:disable=no-name-in-module, import-error
from distutils.core import setup
from setuptools import find_packages

setup(
    name='IXWSAuth',
    version='0.2.0',
    author='Infoxchanhe Australia dev team',
    author_email='devs@infoxchange.net.au',
    packages=find_packages(),
    url='http://pypi.python.org/pypi/IXWSAuth/',
    license='MIT',
    description='Authentication libraries for IX web services',
    long_description=open('README').read(),
    install_requires=(
        'Django >= 1.4.0',
        'IXDjango >= 0.2.1',
        'future',
    ),
    tests_require=(
        'aloe',
        'django-tastypie',
        'djangorestframework',
        'mock',
        'pep8',
        'pylint',
        'pylint-mccabe',
    )
)
